#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum

from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.resources import ResourceType


class TransitionType(Enum):
    """
    Types of fade-in and fade-out transition.
    """

    INSTANT = "instant"
    FADE = "fade"


class Channel(Enum):
    """
    Channel name mappings to the Liqidsoap channel/source IDs.
    """

    QUEUE_A = "in_filesystem_0"
    QUEUE_B = "in_filesystem_1"
    HTTP_A = "in_http_0"
    HTTP_B = "in_http_1"
    LIVE_0 = "linein_0"
    LIVE_1 = "linein_1"
    LIVE_2 = "linein_2"
    LIVE_3 = "linein_3"
    LIVE_4 = "linein_4"
    FALLBACK_QUEUE_A = "in_fallback_scheduled_0"
    FALLBACK_QUEUE_B = "in_fallback_scheduled_1"
    FALLBACK_FOLDER = "fallback_folder"
    FALLBACK_PLAYLIST = "fallback_playlist"

    def __str__(self):
        return str(self.value)


class ChannelResolver:
    """
    Helpers for resolving channel enumerations.
    """

    @staticmethod
    def channel_by_string(channel: str):
        """
        Returns the channel enum for a given channel string from Liquidsoap.
        """
        if not channel:
            return None

        if channel == Channel.QUEUE_A.value:
            return Channel.QUEUE_A
        elif channel == Channel.QUEUE_B.value:
            return Channel.QUEUE_B
        elif channel == Channel.HTTP_A.value:
            return Channel.HTTP_A
        elif channel == Channel.HTTP_B.value:
            return Channel.HTTP_B
        elif channel == Channel.LIVE_0.value:
            return Channel.LIVE_0
        elif channel == Channel.LIVE_1.value:
            return Channel.LIVE_1
        elif channel == Channel.LIVE_2.value:
            return Channel.LIVE_2
        elif channel == Channel.LIVE_3.value:
            return Channel.LIVE_3
        elif channel == Channel.LIVE_4.value:
            return Channel.LIVE_4
        elif channel == Channel.FALLBACK_QUEUE_A.value:
            return Channel.FALLBACK_QUEUE_A
        elif channel == Channel.FALLBACK_QUEUE_B.value:
            return Channel.FALLBACK_QUEUE_B
        elif channel == Channel.FALLBACK_FOLDER.value:
            return Channel.FALLBACK_FOLDER
        elif channel == Channel.FALLBACK_PLAYLIST.value:
            return Channel.FALLBACK_PLAYLIST
        else:
            return None

    @staticmethod
    def live_channel_for_resource(channel: str):
        """
        Returns the channel enum for a given live channel string from Tank.
        """
        if not channel:
            return None
        channel = "linein_" + channel.split("line://")[1]

        if channel == Channel.LIVE_0.value:
            return Channel.LIVE_0
        elif channel == Channel.LIVE_1.value:
            return Channel.LIVE_1
        elif channel == Channel.LIVE_2.value:
            return Channel.LIVE_2
        elif channel == Channel.LIVE_3.value:
            return Channel.LIVE_3
        elif channel == Channel.LIVE_4.value:
            return Channel.LIVE_4
        else:
            return None


class ChannelType(Enum):
    """
    Engine channel types mapped to `Entry` source types.
    """

    QUEUE = {"id": "fs", "numeric": 0, "channels": [Channel.QUEUE_A, Channel.QUEUE_B]}
    HTTP = {"id": "http", "numeric": 1, "channels": [Channel.HTTP_A, Channel.HTTP_B]}
    LIVE = {
        "id": "live",
        "numeric": 3,
        "channels": [
            Channel.LIVE_0,
            Channel.LIVE_1,
            Channel.LIVE_2,
            Channel.LIVE_3,
            Channel.LIVE_4,
        ],
    }
    FALLBACK_QUEUE = {
        "id": "fallback_queue",
        "numeric": 4,
        "channels": [Channel.FALLBACK_QUEUE_A, Channel.FALLBACK_QUEUE_B],
    }
    FALLBACK_POOL = {
        "id": "fallback_pool",
        "numeric": 5,
        "channels": [Channel.FALLBACK_FOLDER, Channel.FALLBACK_PLAYLIST],
    }

    @property
    def channels(self):
        return self.value["channels"]

    @property
    def numeric(self):
        return self.value["numeric"]

    def __str__(self):
        return str(self.value["id"])


class EntryPlayState(Enum):
    UNKNOWN = "unknown"
    LOADING = "loading"
    READY = "ready_to_play"
    PLAYING = "playing"
    FINISHED = "finished"


class LiquidsoapResponse(Enum):
    # There are some weird variations of responses coming
    SUCCESS = ["Done", "Done!", "Donee!", "OK"]
    STREAM_STATUS_POLLING = "polling"
    STREAM_STATUS_STOPPED = "stopped"
    STREAM_STATUS_CONNECTED = "connected"


class ChannelRouter:
    """
    Wires source types with channels and channel-types.
    """

    config = None
    logger = None
    resource_mapping = None
    active_channels = None

    def __init__(self, config, logger):
        """
        Constructor

        Args:
            config (AuraConfig):    The configuration
            logger (Logger):        The logger
        """
        self.config = config
        self.logger = logger

        self.resource_mapping = {
            ResourceType.FILE: ChannelType.QUEUE,
            ResourceType.STREAM_HTTP: ChannelType.HTTP,
            ResourceType.LINE: ChannelType.LIVE,
            ResourceType.PLAYLIST: ChannelType.QUEUE,
            ResourceType.POOL: ChannelType.QUEUE,
        }

        self.active_channels = {
            ChannelType.QUEUE: Channel.QUEUE_A,
            ChannelType.FALLBACK_QUEUE: Channel.FALLBACK_QUEUE_A,
            ChannelType.HTTP: Channel.HTTP_A,
            ChannelType.LIVE: Channel.LIVE_0,
        }

    def set_active(self, channel_type, channel):
        """
        Set the channel for the given resource type active
        """
        self.active_channels[channel_type] = channel

    def get_active(self, channel_type):
        """
        Retrieves the active channel for the given resource type
        """
        return self.active_channels[channel_type]

    def type_of_channel(self, channel):
        """
        Retrieves a `ChannelType` for the given `Channel`.
        """
        if channel in ChannelType.QUEUE.channels:
            return ChannelType.QUEUE
        elif channel in ChannelType.FALLBACK_QUEUE.channels:
            return ChannelType.FALLBACK_QUEUE
        elif channel in ChannelType.HTTP.channels:
            return ChannelType.HTTP
        elif channel in ChannelType.LIVE.channels:
            return ChannelType.LIVE
        else:
            return None

    def is_any_queue(self, channel):
        """
        Evaluates if the channel is of any queue type.
        """
        if (
            self.type_of_channel(channel) == ChannelType.QUEUE
            or self.type_of_channel(channel) == ChannelType.FALLBACK_QUEUE
        ):
            return True
        return False

    def type_for_resource(self, resource_type):
        """
        Retrieves a `ChannelType` for the given `ResourceType`.

        Only default mappings can be evaluatated. Custom variations
        like fallback channels are not respected.
        """
        return self.resource_mapping.get(resource_type)

    def channel_swap(self, channel_type):
        """
        Returns the currently inactive channel for a given type. For example if currently some
        file on channel QUEUE A is playing, the channel QUEUE B is returned for being used
        to queue new entries.

        Args:
            entry_type (ResourceType): The resource type such es file, stream or live source

        Returns:
            (Channel, Channel): The previous and new channel
        """
        previous_channel = self.active_channels[channel_type]
        new_channel = None
        msg = None

        if channel_type == ChannelType.QUEUE:

            if previous_channel == Channel.QUEUE_A:
                new_channel = Channel.QUEUE_B
                msg = "Swapped queue channel from A > B"
            else:
                new_channel = Channel.QUEUE_A
                msg = "Swapped queue channel from B > A"

        elif channel_type == ChannelType.FALLBACK_QUEUE:

            if previous_channel == Channel.FALLBACK_QUEUE_A:
                new_channel = Channel.FALLBACK_QUEUE_B
                msg = "Swapped fallback queue channel from A > B"
            else:
                new_channel = Channel.FALLBACK_QUEUE_A
                msg = "Swapped fallback channel from B > A"

        elif channel_type == ChannelType.HTTP:

            if previous_channel == Channel.HTTP_A:
                new_channel = Channel.HTTP_B
                msg = "Swapped HTTP Stream channel from A > B"
            else:
                new_channel = Channel.HTTP_A
                msg = "Swapped HTTP Stream channel from B > A"

        else:
            self.logger.warning(
                SU.red(f"No channel to swap - invalid entry_type '{channel_type}'")
            )

        if msg:
            self.logger.info(SU.pink(msg))
        return (previous_channel, new_channel)
